import base64
import re
import subprocess
import urllib2

from django.conf import settings
from django.http import HttpResponse

from lxml import etree

from android_build.frontend.views import (
    can_edit,
    is_builder,
    is_official_builder,
    )


auth_headers = {
    'Authorization': 'Basic %s' % (
        base64.encodestring('%s:%s' % (
            settings.FRONTEND_JENKINS_USER,
            settings.FRONTEND_JENKINS_PASSWORD))[:-1],),
    }

valid_build_name = re.compile(r"^[a-zA-Z0-9][-a-zA-Z0-9+.]+$")


def _authJenkins(jenkins_path, data=None, extra_headers=None):
    """Make an authenticated request to jenkins.

    @param jenkins_path: The path on the Jenkins instance to make the request
        to.
    @param data: Data to include in the request (if this is not None the
        request will be a POST).
    @param extra_headers: A dictionary of extra headers that will passed in
        addition to Authorization.
    @raises urllib2.HTTPError: If the response is not a HTTP 200.
    @returns: the body of the response.
    """
    headers = auth_headers.copy()
    if extra_headers:
        headers.update(extra_headers)
    req = urllib2.Request(
        settings.SERVER_JENKINS_URL + jenkins_path, data, headers)
    resp = urllib2.urlopen(req)
    return resp.read()


def messageFromJenkinsErrorPage(page_text):
    """Try to extract a reasonable error message from a Jenkins error page.
    """
    # That this xpath works is probably something of an implementation
    # detail...
    tree = etree.XML(page_text)
    return tree.xpath('//p[../a[@name="skip2content"]]')[0].text


def messageFromWinstoneErrorPage(page_text):
    """Try to extract a reasonable error message from a Winstone error page.
    """
    # this is a complete mess of heuristics and doesn't always work very well
    # :(
    proc = subprocess.Popen(
        ['tidy', '-q','-asxhtml'], stdout=subprocess.PIPE,
        stdin=subprocess.PIPE)
    proc.stdin.write(page_text)
    proc.stdin.close()
    tree = etree.XML(proc.stdout.read().replace(
        'xmlns="http://www.w3.org/1999/xhtml"', '', 1))
    cand = tree.xpath('//h1')[0].tail[1:]
    if cand != 'Exception:':
        return cand
    else:
        return tree.xpath('//pre')[0].text.splitlines()[1]


def getJobConfig(job_name):
    try:
        return _authJenkins('job/' + job_name + '/config.xml')
    except urllib2.HTTPError, e:
        return HttpResponse(
            messageFromWinstoneErrorPage(e.read()), status=e.code)


def replaceConfigDefault(configXml, newValue):
    doc_root = etree.XML(configXml)
    defaultValue = doc_root.xpath('//defaultValue[../name="CONFIG"]')[0]
    defaultValue.text = base64.encodestring(newValue)
    return etree.tostring(doc_root)


def postConfig(url, configXml):
    try:
        _authJenkins(url, configXml, {'Content-Type': 'text/xml'})
        return None
    except urllib2.HTTPError, e:
        return HttpResponse(
            messageFromJenkinsErrorPage(e.read()),
            status=e.code)


def builder_required(func):
    def decorated(request, *args, **kw):
        if not is_builder(request):
            return HttpResponse("you are not allowed to to that", status=403)
        else:
            return func(request, *args, **kw)
    return decorated


def editor_required(func):
    def decorated(request, *args, **kw):
        build_name = request.POST['build']
        if not can_edit(request, build_name):
            return HttpResponse("you are not allowed to to that", status=403)
        else:
            return func(request, *args, **kw)
    return decorated


def job_name_from_request(request):
    if 'build' in request.POST:
        return request.POST['build'][1:].replace('/', '_', 1)
    else:
        return request.GET['build'][1:].replace('/', '_', 1)


@builder_required
def build_now(request):
    job_name = request.POST['build'][1:].replace('/', '_', 1)
    try:
        _authJenkins('job/' + job_name + '/buildWithParameters?delay=0sec')
    except urllib2.HTTPError, e:
        return HttpResponse(
            messageFromWinstoneErrorPage(e.read()),
            status=e.code)
    else:
        return HttpResponse(status=200)


@editor_required
def delete(request):
    _authJenkins('job/' + job_name_from_request(request) + '/doDelete', {})
    return HttpResponse(status=200)


@editor_required
def edit(request):
    job_name = job_name_from_request(request)
    new_text = request.POST['newText']
    oldConfig = getJobConfig(job_name)
    newConfig = replaceConfigDefault(oldConfig, new_text)
    resp = postConfig('job/' + job_name + '/config.xml', newConfig)
    if resp:
        return resp
    return HttpResponse(status=200)


@builder_required
def new(request):
    try:
        if request.POST.get('official', 'false') == 'true':
            if not is_official_builder(request):
                return HttpResponse(
                    "You are not allowed to do that", status=403)
            ownerName = 'linaro-android'
        else:
            ownerName = request.user.username
        buildName = request.POST['name']
        if not valid_build_name.match(buildName):
            return HttpResponse("Invalid name.", status=400)
        new_job_name = ownerName + '_' + buildName
        xml_base = getJobConfig('blank')
        if isinstance(xml_base, HttpResponse):
            return xml_base
        new_config = replaceConfigDefault(xml_base, request.POST['config'])
        resp = postConfig('createItem?name=' + new_job_name, new_config)
        if resp:
            return resp
        if request.POST.get('buildNow', False):
            _authJenkins('job/' + new_job_name + '/buildWithParameters?delay=0sec')
        return HttpResponse(
            '~%s/%s' % (ownerName, buildName),
            status=200)
    except Exception, e:
        return HttpResponse("Unknown error:" + str(e), status=500)


def is_daily(request):
    job_xml = getJobConfig(job_name_from_request(request))
    if isinstance(job_xml, HttpResponse):
        return job_xml
    # I guess that "hudson" might change to "jenkins" here...
    spec = etree.XML(job_xml).xpath(
        "/project/triggers/hudson.triggers.TimerTrigger/spec")
    if len(spec) != 1:
        return HttpResponse('false')
    return HttpResponse(str(spec[0].text == '@daily').lower())


daily_trigger = etree.XML('''
    <hudson.triggers.TimerTrigger>
      <spec>@daily</spec>
    </hudson.triggers.TimerTrigger>
    ''')


@editor_required
def set_daily(request):
    job_name = job_name_from_request(request)
    job_xml = getJobConfig(job_name)
    if isinstance(job_xml, HttpResponse):
        return job_xml
    # I guess that "hudson" might change to "jenkins" here...
    tree = etree.XML(job_xml)
    triggers = tree.xpath("/project/triggers")
    if request.POST['daily'] == 'true':
        triggers[0][:] = [daily_trigger]
    else:
        triggers[0][:] = []
    resp = postConfig('job/' + job_name + '/config.xml', etree.tostring(tree))
    if resp:
        return resp
    return HttpResponse('foo')
